import org.jdom.*;
import org.jdom.input.*;

import java.util.*;
import java.lang.reflect.*;

/**
 * Make a JDOMFactory for a given set of subclasses where
 * the required factory methods simply construct instances
 * of the subclass.
 * @author rpc
 * @version 1.0
 */

public class FactoryMaker
{
    private HashMap classes = new HashMap();

    {
        classes.put(Element.class,Element.class);
        classes.put(Attribute.class, Attribute.class);
        classes.put(CDATA.class, CDATA.class);
        classes.put(DocType.class, DocType.class);
        classes.put(Document.class, Document.class);
        classes.put(Comment.class, Comment.class);
        classes.put(ProcessingInstruction.class, ProcessingInstruction.class);
        classes.put(EntityRef.class, EntityRef.class);
    }

    /**
     * A default factory maker returns usual JDOM classes
     */

    public FactoryMaker() {
    }

    /**
     * Construct a factory maker with the specified single subclass.
     * @param aClass subclass.
     */

    public FactoryMaker(Class aClass) {
        this(new Class [] { aClass });
    }

    /**
     * Construct a factory maker with a list of subclasses to use.
     * @param classes list of subclasses
     */

    public FactoryMaker(Class [] classes) {
        for (int i = 0; i < classes.length; i++)
        {
            setClass(classes[i]);
        }
    }

    /**
     * Set a subclass of the specified type.
     * @param aClass subclass.
     */

    public void setClass(Class aClass) {
        final Class aClassCopy = aClass;

        while (aClass != null) {
            if (classes.get(aClass) != null) {
                classes.put(aClass, aClassCopy);
                break;
            }
            aClass = aClass.getSuperclass();
        }
    }

    /**
     * Return a factory that will make the previously configured subclasses.
     * @return JDOMFactory implementation.
     */

    public JDOMFactory getFactory() {
        return (JDOMFactory) Proxy.newProxyInstance(
            JDOMFactory.class.getClassLoader(),
            new Class [] { JDOMFactory.class },
            new FactoryInvocationHandler(classes));
    }
}

/**
 * InvocationHandler implmentation that turns factory
 * method invocations into concrete constructor
 * invocations with the same parameters.
 */

class FactoryInvocationHandler implements InvocationHandler
{
    private Map classes;

    /**
     * Create handler with specified class map.
     * @param classes class map.
     */

    public FactoryInvocationHandler(Map classes) {
        this.classes = classes;
    }

    public Object invoke(Object proxy,
                         Method method,
                         Object[] args) throws Throwable {

        Class type = method.getReturnType();
        Class subclassType = (Class) classes.get(type);
        Constructor constructor;
        try {
          constructor = subclassType.getConstructor(method.getParameterTypes());
        }
        catch (NoSuchMethodException e) { // missing constructor, use base class
          constructor = type.getConstructor(method.getParameterTypes());
          System.err.println("Constructor for " + subclassType.getName() + " with params " + method.getParameterTypes() + " not defined; using " + type.getName());
        }
        return constructor.newInstance(args);
    }
}

